/* lsmfix.c
 * Fixing Zeiss LSM files.
 * If the Zeiss software is displaying a colormap when saving the image,
 * the PhotometricInterpretation tag is set to 3 (Palette color). This is
 * wrong. The problem is that the ColorMap tag will contain only 256
 * entries, even if the BitsPerSample is set to 16. libtiff will not open
 * such a file. Additionally, only the first directory contains a ColorMap
 * tag, all others are lacking and therefore invalid to libtiff.
 * This program sets the PhotometricInterpretation tag to 2 in each directory.
 *
 * Copyright Cris Luengo, 24 September 2004.
 * Based on tiffdump.c in libtiff, and contains some things copied from libtiff.
 */

/*
 * Warning! This will not work on 64-bit machines!!!
 * We're expecting  uint32 == unsigned long.
 */
 
#include <stdio.h>
#include <stdlib.h>

/* #include "tiffio.h" */
/* Bunch of stuff copied from tiff.h from libtiff */
#define	TIFF_VERSION	        42
#define	TIFF_BIGENDIAN		0x4d4d
#define	TIFF_LITTLEENDIAN	0x4949
typedef	unsigned short uint16;
typedef	unsigned long uint32;
typedef	struct {
	uint16	tiff_magic;	/* magic number (defines byte order) */
	uint16	tiff_version;	/* TIFF version number */
	uint32	tiff_diroff;	/* byte offset to first directory */
} TIFFHeader;
typedef	struct {
	uint16		tdir_tag;	/* see below */
	uint16		tdir_type;	/* data type; see below */
	uint32		tdir_count;	/* number of items; length in spec */
	uint32		tdir_offset;	/* byte offset to field data */
} TIFFDirEntry;
#define	TIFFTAG_PHOTOMETRIC		262	/* photometric interpretation */
/* end of tiff.h */
typedef long toff_t;    /* file offset; off_t is not defined on all platforms? */

char* appname;
int swapflag;
int bigendian;
int errstatus;
int change;

static void FixFile(FILE*);

int main(int argc, char* argv[]) {
   int ii, one = 1;
   FILE* file;
   int cumulative_errstatus = 0;
   errstatus = 0;
   bigendian = *(char*)&one == 0;
   appname = argv[0];
   
   if (argc < 2) {
      printf("Usage: %s file[.lsm]\n", appname);
      return(-1);
   }
   for (ii=1; ii<argc; ii++) {
      file = fopen(argv[ii], "r+b");
      if (!file) {
         perror(0);
         return (-1);
      }
      swapflag = 0;
      change = 0;
      FixFile(file);
      fclose(file);
      if (change) {
         printf("Fixed the file %s\n", argv[ii]);
      } else {
         printf("Nothing to fix in %s\n", argv[ii]);
      }
      cumulative_errstatus |= errstatus;
      errstatus = 0;
   }
   return (cumulative_errstatus);
}


/* Copied from tif_swab.c to avoid having to link with libtiff */
void TIFFSwabShort(uint16* wp) {
	register unsigned char* cp = (unsigned char*) wp;
	unsigned char t;

	t = cp[1]; cp[1] = cp[0]; cp[0] = t;
}
void TIFFSwabLong(uint32* lp) {
	register unsigned char* cp = (unsigned char*) lp;
	unsigned char t;

	t = cp[3]; cp[3] = cp[0]; cp[0] = t;
	t = cp[2]; cp[2] = cp[1]; cp[1] = t;
}


static toff_t FixDirectory(FILE* file, unsigned int ix, toff_t off)
{
   TIFFDirEntry *dp;
   unsigned int n, ii;
   TIFFDirEntry *dir = 0;
   uint16 dircount;
   uint32 nextdiroff = 0;
   uint32 newvalue = 2;
   if (swapflag)
      TIFFSwabLong(&newvalue);

   if (off == 0)        /* no more directories */
      goto done;
   if (fseeko(file, off, 0) != 0) {
      fprintf(stderr, "Seek error accessing TIFF directory.\n");
      errstatus = -1;
      goto done;
   }
   if (fread(&dircount, sizeof(uint16), 1,  file) != 1) {
      fprintf(stderr, "Could not read directory count.\n");
      errstatus = -1;
      goto done;
   }
   if (swapflag)
      TIFFSwabShort(&dircount);
   dir = (TIFFDirEntry *)malloc(dircount * sizeof(TIFFDirEntry));
   if (dir == NULL) {
      fprintf(stderr, "No space for TIFF directory\n");
      errstatus = -1;
      goto done;
   }
   n = fread(dir, sizeof(*dp), dircount, file);
   if (n != dircount) {
      fprintf(stderr, "Could only read %u of %u entries in directory at offset %#lx\n",
          n, dircount, (unsigned long)off);
      dircount = n;
   }
   /*
   printf("Directory offset %#lx (%d tags)\n", (unsigned long)off, dircount);
   */
   if (fread(&nextdiroff, sizeof(uint32), 1, file) != 1)
      nextdiroff = 0;
   if (swapflag)
      TIFFSwabLong(&nextdiroff);
   for (dp = dir, ii = 0; ii<dircount; ii++, dp++) {
      if (swapflag)
         TIFFSwabShort(&dp->tdir_tag);
      if (dp->tdir_tag == TIFFTAG_PHOTOMETRIC) {
         /* Found our target tag. Check to see if it's 3 and convert to 2. */
         /* We know (not checking!): tdir_type = SHORT && tdir_count = 1 */
         if (swapflag)
            TIFFSwabLong(&dp->tdir_offset);
         dp->tdir_offset &= 0xffff;
         if (dp->tdir_offset == 3) {
            /* Find the location in the file where this comes from and write a '2' in there! */
            /*
            printf("Changing a value!!!\n");
            */
            if (fseeko(file, off + sizeof(uint16) + ii*sizeof(TIFFDirEntry) + 2*sizeof(uint16)+sizeof(uint32), 0) != 0) {
               fprintf(stderr, "Seek error accessing TIFF tag for writing.\n");
               errstatus = -1;
               goto done;
            }
            if (fwrite(&newvalue, sizeof(uint32), 1,  file) != 1) {
               fprintf(stderr, "Could not write new PhotometricInterpretation value.\n");
               errstatus = -1;
               goto done;
            }
            change = 1;
         }
      }
   }     
done:
   if (dir)
      free(dir);
   return (nextdiroff);
}


static void FixFile(FILE* file) {
   TIFFHeader hdr;
   unsigned int ii;
   toff_t diroff;

   if (fread(&hdr, sizeof (hdr), 1,  file) != 1) {
      fprintf(stderr, "Could not read header.\n");
      errstatus = -1;
      return;
   }
   /*
    * Setup the byte order handling.
    */
   if (hdr.tiff_magic != TIFF_BIGENDIAN && hdr.tiff_magic != TIFF_LITTLEENDIAN) {
      fprintf(stderr, "Not a TIFF file, bad magic number %u (%#x)\n", hdr.tiff_magic, hdr.tiff_magic);
      errstatus = -1;
      return;
   }
   if (hdr.tiff_magic == TIFF_BIGENDIAN) {
      swapflag = !bigendian;
   } else {
      swapflag = bigendian;
   }
   /*
    * Swap header if required.
    */
   if (swapflag) {
      TIFFSwabShort(&hdr.tiff_version);
      TIFFSwabLong(&hdr.tiff_diroff);
   }
   /*
    * Now check version (if needed, it's been byte-swapped).
    * Note that this isn't actually a version number, it's a
    * magic number that doesn't change (stupid).
    */
   if (hdr.tiff_version != TIFF_VERSION) {
      fprintf(stderr, "Not a TIFF file, bad version number %u (%#x)\n", hdr.tiff_version, hdr.tiff_version);
      errstatus = -1;
      return;
   }
   /*
    * Do the thing in each directory.
    */
   diroff = hdr.tiff_diroff;
   for (ii = 0; diroff != 0; ii++) {
      diroff = FixDirectory(file, ii, diroff);
      if (errstatus != 0) return;
   }
}
